/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: acquire_data.cpp 32074 2010-02-25 20:16:36Z rusu $
 *
 */

/**
@mainpage

\author Radu Bogdan Rusu

@b acquire_data controls a Gazebo virtual stereo camera to trigger snapshots for object data acquisition.
**/

#include <Eigen/Core>
#include <Eigen/LU>
#include <Eigen/Geometry>

#include <boost/filesystem.hpp> 
#include <boost/thread.hpp> 
#include <boost/make_shared.hpp> 
#include <ros/ros.h>
#include <ros/callback_queue.h>
#include <rosbag/bag.h>
#include <nav_msgs/Odometry.h>
#include <tf/tfMessage.h>
#include <roslib/Clock.h>
// PointCloud
#include <sensor_msgs/PointCloud.h>
#include <sensor_msgs/PointCloud2.h>
// Image
#include <image_transport/image_transport.h>
#include <image_transport/subscriber_filter.h>
#include <message_filters/subscriber.h>
#include <message_filters/time_synchronizer.h>
// Interface to model database
#include <model_database/postgresql_model_database.h>
#include <model_database/database_original_model.h>
// Interface to Gazebo
#include <gazebo/DeleteModel.h>
#include <gazebo/SpawnModel.h>
#include <gazebo/SetModelState.h>
// Interface to VTK
#include <vtkPlatonicSolidSource.h>
#include <vtkSmartPointer.h>
#include <vtkLoopSubdivisionFilter.h>
#include <vtkTransform.h>
#include <vtkPolyData.h>
#include <vtkTransformPolyDataFilter.h>
#include <vtkMath.h>
#include <vtkCellArray.h>

using namespace std;
using namespace boost;
using namespace ros;
using namespace roslib;
using namespace tf;
using namespace sensor_msgs;
using namespace nav_msgs;
using namespace image_transport;
using namespace model_database;

class ObjectDataAcquisitionGazebo
{
  // Basic point structure
  struct PointXYZRGBA
  {
    float x;
    float y;
    float z;
    uint32_t rgba;
  };

  typedef sensor_msgs::PointCloud2 PointCloud;

  private:
    // ROS
    NodeHandle nh_;
    rosbag::Bag bag_;
    ros::Subscriber tf_sub_, clock_sub_;
    ServiceClient gazebo_model_remove_serv_, gazebo_model_spawn_serv_, gazebo_set_pose_serv_;
    CallbackQueue tf_queue_, clock_queue_;
    thread tf_thread_, clock_thread_;
    mutex m_mutex_;

    // Images
    ImageTransport it_;
    SubscriberFilter image_sub_l_, image_sub_r_;
    message_filters::Subscriber<CameraInfo> info_sub_l_, info_sub_r_;
    message_filters::TimeSynchronizer<Image, CameraInfo, Image, CameraInfo> sync_;
   
    // Virtual camera positions
    sensor_msgs::PointCloud camera_positions_;
    // Image pointers
    ImageConstPtr raw_image_l_, raw_image_r_;
    CameraInfoConstPtr cam_info_l_, cam_info_r_;
 
    // Model database
    shared_ptr<PostgresqlModelDatabase> database_;
    ptr_vector<DatabaseOriginalModel> models_;
    std::string root_path_;

    // Topics
    string t_camera_base_, t_ptu_base_, t_pan_tilt_, t_image_raw_l_, t_image_raw_r_, t_cam_info_l_, t_cam_info_r_;

    // Output BAG file
    string output_file_;

    double scale_factor_;
    bool log_file_open_, init_failed_, alternate_mode_, artificial_texture_, alternate_artificial_texture_;


  public:
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ObjectDataAcquisitionGazebo (int texture) : nh_ ("~"), it_ (nh_), sync_ (3), 
                              raw_image_l_ (make_shared<Image>()), raw_image_r_ (make_shared<Image>()),
                              cam_info_l_ (make_shared<CameraInfo>()), cam_info_r_ (make_shared<CameraInfo>()),
                              t_camera_base_ ("/narrow_stereo"), 
                              log_file_open_ (false), init_failed_ (false), alternate_mode_ (false), 
                              artificial_texture_ (true), alternate_artificial_texture_ (false)
    {
      // Select the correct texturing mode
      switch (texture)
      {
        case 0: artificial_texture_ = false; break;
        case 1: artificial_texture_ = true; break;
        case 2: alternate_artificial_texture_ = true; alternate_mode_ = true; break;
      }

      nh_.param ("scale_factor", scale_factor_, 1.0);
      if (scale_factor_ != 1.0)
        ROS_INFO ("Using a scale factor of %f.", scale_factor_);
      // All these topics need to be BAGed
      t_image_raw_l_     = t_camera_base_ + "/left/image_raw";
      t_image_raw_r_     = t_camera_base_ + "/right/image_raw";
      t_cam_info_l_      = t_camera_base_ + "/left/camera_info";
      t_cam_info_r_      = t_camera_base_ + "/right/camera_info";

      // Subscribe to TF & clock
      SubscribeOptions tf_op    = SubscribeOptions::create<tfMessage>("/tf", 1, bind (&ObjectDataAcquisitionGazebo::tfCB, this, _1), VoidPtr (), &tf_queue_);
      SubscribeOptions clock_op = SubscribeOptions::create<Clock>("/clock", 1, bind (&ObjectDataAcquisitionGazebo::clockCB, this, _1), VoidPtr (), &clock_queue_);
      tf_sub_    = nh_.subscribe (tf_op);
      clock_sub_ = nh_.subscribe (clock_op);

      tf_thread_    = thread (bind (&ObjectDataAcquisitionGazebo::tfThread, this));
      clock_thread_ = thread (bind (&ObjectDataAcquisitionGazebo::clockThread, this));

      //tf_sub_ = nh_.subscribe ("/tf", 1, &ObjectDataAcquisitionGazebo::tfCB, this);
      //clock_sub_ = nh_.subscribe ("/clock", 1, &ObjectDataAcquisitionGazebo::clockCB, this);
      
      // Subscribe to the raw camera data (left, right, info left, info right)
      info_sub_l_.subscribe (nh_, t_cam_info_l_, 1);
      info_sub_r_.subscribe (nh_, t_cam_info_r_, 1);
      image_sub_l_.subscribe (it_, t_image_raw_l_, 1);
      image_sub_r_.subscribe (it_, t_image_raw_r_, 1);
      sync_.connectInput (image_sub_l_, info_sub_l_, image_sub_r_, info_sub_r_);
      sync_.registerCallback (bind (&ObjectDataAcquisitionGazebo::imageCB, this, _1, _2, _3, _4));

      // Gazebo
      gazebo_model_remove_serv_ = nh_.serviceClient<gazebo::DeleteModel>("/gazebo/delete_model");
      gazebo_model_spawn_serv_  = nh_.serviceClient<gazebo::SpawnModel>("/gazebo/spawn_gazebo_model");
      gazebo_set_pose_serv_ = nh_.serviceClient<gazebo::SetModelState>("/gazebo/set_model_state");
      // wait for services to become available
      gazebo_model_remove_serv_.waitForExistence();
      gazebo_model_spawn_serv_.waitForExistence();
      gazebo_set_pose_serv_.waitForExistence();

      // Init the model database
      database_ = make_shared<PostgresqlModelDatabase> ("wgs36", "5432", "willow", "willow", "models");
      if (!database_->isConnected ())
      {
        ROS_ERROR ("Failed to connect to database");
        init_failed_ = true;
        return;
      }

      if (!database_->getOriginalModelsList (models_))
      {
        ROS_ERROR ("Could not get models from database. Die in horrible pain.");
        init_failed_ = true;
        return;
      }
      ROS_INFO ("Number of models in the database: %d", (int)models_.size ());
      if (!database_->getModelRoot (root_path_))
      {
        ROS_ERROR ("Failed to get root path from database");
        init_failed_ = true;
        return;
      }
      
      if (artificial_texture_)
        ROS_INFO ("Artificial texture enabled.");
      else
        ROS_INFO ("Using the objects' original texture.");

      // Estimate all the camera positions on a virtual sphere (80 positions) 
      computeSphereTessellation (1, scale_factor_, camera_positions_);
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
      clockThread ()
    {
      static const double timeout = 0.01;
      while (nh_.ok ())
        clock_queue_.callAvailable (WallDuration (timeout));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
      tfThread ()
    {
      static const double timeout = 0.01;
      while (nh_.ok ())
        tf_queue_.callAvailable (WallDuration (timeout));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    virtual ~ObjectDataAcquisitionGazebo ()
    {
      // Clear and disable the queues
      tf_queue_.clear ();    tf_queue_.disable ();
      clock_queue_.clear (); clock_queue_.disable ();
      // Shut down ROS
      requestShutdown ();
      // Join the threads
      tf_thread_.join ();
      clock_thread_.join ();

      if (init_failed_)
        return;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Stereo image callback 
      * \param raw_image_l the raw left image
      * \param cam_info_l the left camera info
      * \param raw_image_r the raw right image
      * \param cam_info_r the right camera info
      */
    void
      imageCB (const ImageConstPtr& raw_image_l, const CameraInfoConstPtr& cam_info_l, 
               const ImageConstPtr& raw_image_r, const CameraInfoConstPtr& cam_info_r)
    {
      raw_image_l_ = raw_image_l;
      raw_image_r_ = raw_image_r;
      cam_info_l_  = cam_info_l;
      cam_info_r_  = cam_info_r;
      ROS_DEBUG ("Image pair (%dx%d) received. [%s - %s]", raw_image_l_->width, raw_image_l_->height, t_image_raw_l_.c_str (), t_image_raw_r_.c_str ());
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief TF callback
      * \param msg the TF message
      */ 
    void
      tfCB (const tfMessageConstPtr &msg)
    {
      mutex::scoped_lock lock (m_mutex_);
      if (!log_file_open_) return;
      bag_.write ("tf", Time::now (), msg);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Clock callback
      * \param msg the clock message
      */ 
    void
      clockCB (const ClockConstPtr &msg)
    {
      mutex::scoped_lock lock (m_mutex_);
      if (!log_file_open_) return;
      bag_.write ("/clock", Time::now (), msg);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Estimate all camera positions on a tessellated Icosahedron for a given subdivision level
      * \param level the level of subdivision (see \a vtkLoopSubdivisionFilter)
      * \param scale_factor multiply the radius of the resultant sphere with this factor
      * \param sphere the resultant centers of the triangles on the tessellated sphere as a PointCloud
      */
    void
      computeSphereTessellation (int level, double scale_factor, sensor_msgs::PointCloud &sphere)
    {
      // Create a Icosahedron
      vtkSmartPointer<vtkPlatonicSolidSource> icosa = vtkSmartPointer<vtkPlatonicSolidSource>::New ();
      icosa->SetSolidTypeToIcosahedron ();
      
      // Tesselize the source sphere/icosahedron (subdivide each triangular face
      // of the icosahedron into four smaller triangles)
      vtkTransformPolyDataFilter* tf = vtkTransformPolyDataFilter::New ();
      
      vtkSmartPointer<vtkLoopSubdivisionFilter> subdivide = vtkSmartPointer<vtkLoopSubdivisionFilter>::New ();
      subdivide->SetNumberOfSubdivisions (level);
      subdivide->SetInputConnection (icosa->GetOutputPort ());

      // Compute the scale distance 
      vtkPolyData* result = subdivide->GetOutput ();
      result->Update ();
      double distance_total = 0;
      double center[3] = {.0, .0, .0};
      for (int i = 0; i < result->GetNumberOfPoints (); ++i)
      {
        double p[3];
        result->GetPoint (i, p);
        distance_total += 1 / vtkMath::Distance2BetweenPoints (center, p);
      }
      distance_total /= result->GetNumberOfPoints ();
      double scale = sqrt (distance_total);
      ROS_INFO ("Radius before scale: %f", scale);
      
      tf->SetInputConnection (subdivide->GetOutputPort ());

      // Apply the scale
      vtkSmartPointer<vtkTransform> t = vtkSmartPointer<vtkTransform>::New ();
      scale *= scale_factor;
      t->Identity (); t->Scale (scale, scale, scale);

      tf->SetTransform (t);
          
      result = tf->GetOutput ();
      result->Update ();

      // Re-estimate the radius
      distance_total = 0;
      center[0] = center[1] = center[2] = 0;
      for (int i = 0; i < result->GetNumberOfPoints (); ++i)
      {
        double p[3];
        result->GetPoint (i, p);
        distance_total += 1 / vtkMath::Distance2BetweenPoints (center, p);
      }
      distance_total /= result->GetNumberOfPoints ();
      ROS_INFO ("Radius after scale: %f", sqrt (distance_total));

      // Get the polygons
      vtkCellArray *polys = result->GetPolys ();

      // Get the centroids
      polys->InitTraversal ();
      int res = 0;
      double p[3];
      geometry_msgs::Point32 pt;
      do
      {
        vtkIdType npts;
        vtkIdType *pts;
        res = polys->GetNextCell (npts, pts);

        // Compute the centroid of each cell
        if (res != 0)
        {
          pt.x = pt.y = pt.z = 0;
          // Iterate over all points
          for (int cp = 0; cp < npts; ++cp)
          {
            result->GetPoint (pts[cp], p);
            pt.x += p[0]; pt.y += p[1]; pt.z += p[2];
          }
          pt.x /= npts; pt.y /= npts; pt.z /= npts;
          sphere.points.push_back (pt);
        }
      } while (res != 0);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Converts a 3D point sampled on a sphere to a Gazebo camera pose pointing at the origin
      * \param sphere_pt the point on the camera view sphere
      * \param now a timestamp
      * \param odom the resultant Odometry message representing the Gazebo camera pose
      */
    void
      computeGazeboCameraPose (const geometry_msgs::Point32 &pt, const Time &now, Odometry &camera_odom)
    {
      Eigen::Vector3d v (-pt.x, -pt.y, 0);
      Eigen::Vector3d u ( pt.y, -pt.x, 0);
      v.normalize (); u.normalize ();
      Eigen::Vector3d w = v.cross (u);
      if (w.squaredNorm () == 0)
      {
        u[0] = -pt.z; u[1] = 0; u[2] = pt.x;
        w = v.cross (u);
      }

      //std::cout << "debug " << std::endl << w << std::endl;

      // Compute the orientation of the camera
      Eigen::Matrix3d R;
      R << v, u, w;
      Eigen::Quaternion<double> q (R);

      // Set the position
      geometry_msgs::Point cam_pt; cam_pt.x = pt.x; cam_pt.y = pt.y; cam_pt.z = pt.z;
      // Set the orientation
      geometry_msgs::Quaternion quat; quat.x = q.coeffs()[0]; quat.y = q.coeffs ()[1]; quat.z = q.coeffs ()[2]; quat.w = q.coeffs ()[3]; 

      geometry_msgs::Pose pose; pose.position = cam_pt; pose.orientation = quat;
      array<double, 36> COV;
      Eigen::Map<Eigen::MatrixXd> COVm (&COV[0], 6, 6);
      COVm.setZero ();
      geometry_msgs::PoseWithCovariance pwc; pwc.pose = pose; pwc.covariance = COV;
      camera_odom.header.stamp = now;
      camera_odom.header.frame_id = "/map";
      camera_odom.child_frame_id  = camera_odom.header.frame_id;
      camera_odom.pose = pwc;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Constructs a Gazebo URDF model to be sent containing an object model
      * \param model_path the path to the object model
      * \param artificial_texture set to true if we want artificial texture
      */ 
    string
      assembleModelGazebo (string model_path, bool artificial_texture)
    {
      string model = 
        "<?xml version=\"1.0\" ?>\n"
         "<model:physical name=\"object_model\">\n"
         "<xyz>   0.0    0.0    0.0</xyz>\n"
         "<rpy>   0.0    0.0    0.0</rpy>\n"
         "<static>false</static>\n"
         "<body:trimesh name=\"object_body\">\n"
         "  <turnGravityOff>true</turnGravityOff>\n"
         "  <massMatrix>true</massMatrix>\n"
         "  <mass>1.0</mass>\n"
         "  <ixx>0.1</ixx>\n"
         "  <ixy>0.0</ixy>\n"
         "  <ixz>0.0</ixz>\n"
         "  <iyy>0.1</iyy>\n"
         "  <iyz>0.0</iyz>\n"
         "  <izz>0.1</izz>\n"
         "  <cx>0.0</cx>\n"
         "  <cy>0.0</cy>\n"
         "  <cz>0.05</cz>\n"
         "  <geom:trimesh name=\"object_geom\">\n"
         "    <xyz>0 0 0</xyz>\n"
         "    <laserRetro>2000.0</laserRetro>\n"
         "    <kp>100000000.0</kp>\n"
         "    <kd>1.0</kd>\n"
         "    <scale>0.001 0.001 0.001</scale>\n"
         "    <mesh>"+ model_path + "</mesh>\n"
         "    <centerMesh>aabb_center</centerMesh>\n"
         "    <genTexCoord>true</genTexCoord>\n"
         "    <visual>\n"
         "      <xyz>0 0 0</xyz>\n"
         "      <scale>0.001 0.001 0.001</scale>\n";
      if (artificial_texture)
        model += string ("      <material>GazeboWorlds/RGBNoise</material>\n");
      model += string (
         "      <mesh>"+ model_path + "</mesh>\n"
         "    </visual>\n"
         "  </geom:trimesh>\n"
         "</body:trimesh>\n"
       "</model:physical>");
      return (model);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Records a set of global messages and topics for the current given timestamp
      * \param camera_odom the Gazebo virtual camera pose
      * \param now the current ros::Time
      */ 
    void
      recordTopics (Odometry &camera_odom, const Time &now)
    {
      // Check if we already got data
      while ( (raw_image_l_->data.size () == 0) && nh_.ok () )
      {
        ROS_WARN ("No data received yet. Waiting...");
        spinOnce ();
      }

      // wait until we get new data
      Time t_i_img, t_img = raw_image_l_->header.stamp; 
      do
      {
        // Spin 
        spinOnce ();
        // Get current timestamps
        t_i_img = raw_image_l_->header.stamp;

        // If all timestamps are higher than now, we're good
        if (t_i_img >= now)
          break;

        // Else, warn
        if (t_i_img != t_img)
        {
          ROS_DEBUG ("Image received [%s], but timestamp (%f) is older than now (%f)", t_image_raw_l_.c_str (), t_i_img.toSec (), now.toSec ());
          t_img = t_i_img;
        }
      }
      while ( (t_i_img < now) && nh_.ok ());
    
      if (!log_file_open_) return;
      {
        mutex::scoped_lock lock (m_mutex_);
        bag_.write (t_image_raw_l_, now, raw_image_l_);
        bag_.write (t_image_raw_r_, now, raw_image_r_);
        bag_.write (t_cam_info_l_,  now, cam_info_l_);
        bag_.write (t_cam_info_r_,  now, cam_info_r_);
        // Assemble the pose of the camera
        camera_odom.header.stamp = now;
        bag_.write ("/stereo_pose_ground_truth", now, camera_odom);
      }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
      spin ()
    {
      if (init_failed_)
        return;

      // For all the models in the database
      for (size_t cm = 0; cm < models_.size (); cm++)
      {
        if (!database_->loadFromDatabase (&models_[cm].concave_filled_))
        {
          ROS_WARN ("Model %d does not have concave fill set.", models_[cm].id_.get ());
          continue;
        }
        if (models_[cm].concave_filled_.get ())
          continue;
        // Get the model
        string model_path = models_[cm].geometry_path_x3d_.get ();
        if (model_path.empty ())
          continue;
        string path = root_path_ + model_path;

        // X3D to DAE
        path.replace (path.find ("/x3d/"), 5, "/dae2/");
        path = path.substr (0, path.find_last_of (".")) + ".dae" ;

        if (!boost::filesystem::exists (path))
          continue;
        ROS_INFO ("Loading dae model %d (%s): %s", models_[cm].id_.get (), models_[cm].model_.get ().c_str (), path.c_str ());

        output_file_ = path.substr (path.find_last_of ("/") + 1);
        output_file_.replace (output_file_.find (".dae"), 4, ".bag");
        // Do not attempt to overwrite the bag file
        if (boost::filesystem::exists (output_file_))
          continue;

        // Init the BAG recorder
        {
          mutex::scoped_lock lock (m_mutex_);
          try
          {
            bag_.open (output_file_, rosbag::bagmode::Read | rosbag::bagmode::Write);
            log_file_open_ = true;
          }
          catch (rosbag::BagException)
          {
            ROS_FATAL ("Could not open output file: %s", output_file_.c_str ());
            init_failed_ = true;
          }
        }

        // Call the Gazebo service request to remove the existing model (if any)
        gazebo::DeleteModel dm; dm.request.model_name = "object_model";
        gazebo_model_remove_serv_.call (dm);

        // Assemble the model to be sent to Gazebo
        string gazebo_model = assembleModelGazebo (path, artificial_texture_);

        // Call the Gazebo service request to add the current model
        gazebo::SpawnModel sm; 
        sm.request.model_name = "object_model"; 
        sm.request.model_xml = gazebo_model;
        sm.request.robot_namespace = "/";
        sm.request.initial_pose.position.x = 1.0; // center of table is shifted x by 1
        sm.request.initial_pose.position.z = 0.585; // table top level
        gazebo_model_spawn_serv_.call (sm);
 
        // Go along a view sphere and set the Gazebo camera position and orientation
        int cnt = 0;
        for (size_t cp = 0; cp < camera_positions_.points.size (); ++cp)
        {
          // Estimate the correct camera pose
          Odometry camera_odom;
          computeGazeboCameraPose (camera_positions_.points[cp], Time::now (), camera_odom);

          // If alternate mode is enabled
          if (alternate_mode_)
          {
/*            // Remove the current model
            gazebo::DeleteModel dm; dm.request.model_name = "object_model";
            gazebo_model_remove_serv_.call (dm);

            // Assemble the model to be sent to Gazebo
            string gazebo_model = assembleModelGazebo (path, alternate_artificial_texture_);

            // Call the Gazebo service request to add the current model
            gazebo::SpawnModel sm; 
            sm.request.model.model_name = "object_model"; 
            sm.request.model.robot_model = gazebo_model; sm.request.model.xml_type = 1; sm.request.model.robot_namespace = "/";
            gazebo_model_spawn_serv_.call (sm);

            alternate_artificial_texture_ = !alternate_artificial_texture_;*/
          }

          // Actually change the Gazebo camera pose
          double x = camera_odom.pose.pose.position.x;
          double y = camera_odom.pose.pose.position.y;
          double r = std::max(0.0001,sqrt(x*x+y*y));
          camera_odom.pose.pose.position.x = (r+1.2)*x/r + 1.0; // shift center forward by 1.0, map from r=[0,1] to r=[1.2,2.2]
          camera_odom.pose.pose.position.y = (r+1.2)*y/r;
          camera_odom.pose.pose.position.z = -0.0055; // pr2 ground interpenetration due to weight
          camera_odom.header.frame_id = "world";
          //ROS_INFO("debug: %f %f %f",x,y,r);
          gazebo::SetModelState sp;
          sp.request.model_state.model_name = "pr2";
          sp.request.model_state.pose = camera_odom.pose.pose;
          gazebo_set_pose_serv_.call (sp);
          // Sleep to allow the camera to stabilize
          ros::Duration (0.5).sleep ();
          Time now = Time::now ();     // get the current time

          recordTopics (camera_odom, now);
          ROS_INFO ("Saving camera images, camera info, and point cloud [%d]", cnt++);
          // Sleep to allow TF to stabilize
          ros::Duration (0.5).sleep ();
        }

        ROS_INFO ("Closing the BAG file...");
        {
          log_file_open_ = false;
          mutex::scoped_lock lock (m_mutex_);
          bag_.close ();
        }
      }
    }

};

/* ---[ */
int
  main (int argc, char** argv)
{
  ROS_INFO ("Syntax is: %s [artificial_texture = 0/1/2 (0 = disabled, use the real object texture / 1 = enabled* / 2 = alternate mode, save both)]", argv[0]);

  int param;
  if (argc > 1)
    param = atoi (argv[1]);
  else
    param = 1;
  // ROS init
  init (argc, argv, "object_data_acquisition_gazebo");

  ObjectDataAcquisitionGazebo p (param);
  p.spin ();

  return (0);
}

