/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: acquire_data.cpp 32074 2010-02-25 20:16:36Z rusu $
 *
 */

/**
@mainpage

\author Radu Bogdan Rusu

@b acquire_data controls a Gazebo virtual stereo camera to trigger snapshots for object data acquisition.
**/

#include <Eigen/Core>
#include <Eigen/LU>
#include <Eigen/Geometry>

#include <boost/filesystem.hpp> 
#include <boost/thread.hpp> 
#include <boost/make_shared.hpp> 
#include <ros/ros.h>
#include <ros/callback_queue.h>
#include <nav_msgs/Odometry.h>
#include <tf/tfMessage.h>
#include <roslib/Clock.h>
// Image
#include <image_transport/image_transport.h>
#include <image_transport/subscriber_filter.h>
#include <message_filters/subscriber.h>
#include <message_filters/time_synchronizer.h>
// Interface to model database
#include <model_database/postgresql_model_database.h>
#include <model_database/database_original_model.h>
// Interface to Gazebo
#include <gazebo_plugins/DeleteModel.h>
#include <gazebo_plugins/SpawnModel.h>
#include <gazebo_plugins/SetPose.h>

using namespace std;
using namespace boost;
using namespace ros;
using namespace roslib;
using namespace tf;
using namespace sensor_msgs;
using namespace nav_msgs;
using namespace image_transport;
using namespace model_database;
using namespace gazebo_plugins;

class ObjectDataAcquisitionGazebo
{
  private:
    // ROS
    NodeHandle nh_;
    ServiceClient gazebo_model_remove_serv_, gazebo_model_spawn_serv_, gazebo_camera_set_pose_serv_;
    CallbackQueue tf_queue_, clock_queue_;
    thread tf_thread_, clock_thread_;
    mutex m_mutex_;

    // Model database
    shared_ptr<PostgresqlModelDatabase> database_;
    std::string model_dae_path_;

    // Topics
    string t_camera_base_, t_ptu_base_, t_pan_tilt_, t_image_raw_l_, t_image_raw_r_, t_cam_info_l_, t_cam_info_r_;

    // Output BAG file
    string output_file_;

    double scale_factor_;
    bool log_file_open_, init_failed_, alternate_mode_, artificial_texture_, alternate_artificial_texture_;


  public:
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ObjectDataAcquisitionGazebo (int texture, int model_id) : nh_ ("~"),
							      log_file_open_ (false), 
							      init_failed_ (false), alternate_mode_ (false), 
							      artificial_texture_ (true), 
							      alternate_artificial_texture_ (false)
    {
      // Select the correct texturing mode
      switch (texture)
      {
        case 0: artificial_texture_ = false; break;
        case 1: artificial_texture_ = true; break;
        case 2: alternate_artificial_texture_ = true; alternate_mode_ = true; break;
      }

      nh_.param ("scale_factor", scale_factor_, 1.0);
      if (scale_factor_ != 1.0)
        ROS_INFO ("Using a scale factor of %f.", scale_factor_);

      tf_thread_    = thread (bind (&ObjectDataAcquisitionGazebo::tfThread, this));
      clock_thread_ = thread (bind (&ObjectDataAcquisitionGazebo::clockThread, this));

      // Gazebo
      gazebo_model_remove_serv_ = nh_.serviceClient<DeleteModel>("/delete_model");
      gazebo_model_spawn_serv_  = nh_.serviceClient<SpawnModel>("/spawn_model");
      gazebo_camera_set_pose_serv_ = nh_.serviceClient<SetPose>("/set_pr2_pose");

      gazebo_model_remove_serv_.waitForExistence();
      gazebo_model_spawn_serv_.waitForExistence();

      // Init the model database
      database_ = make_shared<PostgresqlModelDatabase> ("wgs36", "5432", "willow", "willow", "models");
      if (!database_->isConnected ())
      {
        ROS_ERROR ("Failed to connect to database");
        init_failed_ = true;
        return;
      }
      std::string root_path;
      if (!database_->getModelRoot (root_path))
      {
        ROS_ERROR ("Failed to get root path from database");
        init_failed_ = true;
        return;
      }
      model_database::DatabaseOriginalModel db_model;
      db_model.id_.get() = model_id;
      //check if the model exists in the database; look at its name
      if (! database_->loadFromDatabase(&db_model.model_) || db_model.model_.get().empty())
      {
	ROS_ERROR("Model with id %d not found in database", model_id);
	init_failed_ = true;
	return;
      }
      //check if the dae path is there
      if (! database_->loadFromDatabase(&db_model.geometry_path_dae_) || db_model.geometry_path_dae_.get().empty())
      {
	ROS_ERROR("Model with id %d does not have a .dae geometry path", model_id);
	init_failed_ = true;
	return;
      }
      //assemble the complete path and check if file exists
      model_dae_path_ = root_path + db_model.geometry_path_dae_.get();
      if (!boost::filesystem::exists (model_dae_path_))
      {
	ROS_ERROR("DAE file %s not found", model_dae_path_.c_str());
	init_failed_ = true;
	return;
      }
      ROS_INFO("Using dae file %s", model_dae_path_.c_str());
      
      if (artificial_texture_)
        ROS_INFO ("Artificial texture enabled.");
      else
        ROS_INFO ("Using the objects' original texture.");

    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
      clockThread ()
    {
      static const double timeout = 0.01;
      while (nh_.ok ())
        clock_queue_.callAvailable (WallDuration (timeout));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
      tfThread ()
    {
      static const double timeout = 0.01;
      while (nh_.ok ())
        tf_queue_.callAvailable (WallDuration (timeout));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    virtual ~ObjectDataAcquisitionGazebo ()
    {
      // Clear and disable the queues
      tf_queue_.clear ();    tf_queue_.disable ();
      clock_queue_.clear (); clock_queue_.disable ();
      // Shut down ROS
      requestShutdown ();
      // Join the threads
      tf_thread_.join ();
      clock_thread_.join ();

      if (init_failed_)
        return;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** \brief Constructs a Gazebo URDF model to be sent containing an object model
      * \param model_path the path to the object model
      * \param artificial_texture set to true if we want artificial texture
      */ 
    string
      assembleModelGazebo (string model_path, bool artificial_texture)
    {
      string model = 
        "<?xml version=\"1.0\" ?>\n"
         "<model:physical name=\"object_1\">\n"
         "<xyz>   0.0    0.0    0.0</xyz>\n"
         "<rpy>   0.0    0.0    0.0</rpy>\n"
         "<static>false</static>\n"
         "<body:trimesh name=\"object_body\">\n"
         "  <turnGravityOff>false</turnGravityOff>\n"
         "  <massMatrix>true</massMatrix>\n"
         "  <mass>1.0</mass>\n"
         "  <ixx>0.1</ixx>\n"
         "  <ixy>0.0</ixy>\n"
         "  <ixz>0.0</ixz>\n"
         "  <iyy>0.1</iyy>\n"
         "  <iyz>0.0</iyz>\n"
         "  <izz>0.1</izz>\n"
         "  <cx>0.0</cx>\n"
         "  <cy>0.0</cy>\n"
         "  <cz>0.05</cz>\n"
         "  <geom:trimesh name=\"object_geom\">\n"
         "    <xyz>0 0 0</xyz>\n"
         "    <laserRetro>2000.0</laserRetro>\n"
         "    <kp>100000000.0</kp>\n"
         "    <kd>1.0</kd>\n"
         "    <scale>0.001 0.001 0.001</scale>\n"
         "    <mesh>"+ model_path + "</mesh>\n"
         "    <visual>\n"
         "      <xyz>0 0 0</xyz>\n"
         "      <scale>0.001 0.001 0.001</scale>\n";
      if (artificial_texture)
        model += string ("      <material>GazeboWorlds/RGBNoise</material>\n");
      model += string (
         "      <mesh>"+ model_path + "</mesh>\n"
         "    </visual>\n"
         "  </geom:trimesh>\n"
         "</body:trimesh>\n"
       "</model:physical>");
      return (model);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    void
      spin ()
    {
      if (init_failed_) return;

      // Call the Gazebo service request to remove the existing model (if any)
      DeleteModel dm; dm.request.model_name = "object_1";
      gazebo_model_remove_serv_.call (dm);
      
      // Assemble the model to be sent to Gazebo
      string gazebo_model = assembleModelGazebo (model_dae_path_, artificial_texture_);
      
      // Call the Gazebo service request to add the current model
      SpawnModel sm; 
      sm.request.model.model_name = "object_1"; 
      sm.request.model.robot_model = gazebo_model; 
      sm.request.model.xml_type = 1; sm.request.model.robot_namespace = "/";
      sm.request.model.initial_pose.position.x = .7; // center of table is shifted x by 1
      sm.request.model.initial_pose.position.z = 0.55; // table top level
      gazebo_model_spawn_serv_.call (sm);
      ROS_INFO ("wait...");
      usleep(1000000);
    }
};

/* ---[ */
int
  main (int argc, char** argv)
{
  ROS_INFO("Syntax is: %s model_id [artificial_texture = 0/1/2 "
	   "(0 = disabled, use the real object texture / 1 = enabled* / 2 = alternate mode, save both)] ", argv[0]);

  
  int model_i;
  if (argc > 1) model_i = atoi (argv[1]);
  else 
  {
    ROS_ERROR("Can not spawn database object: model_id not specified");
    return (0);
  }

  int param = 1;
  if (argc > 2) param = atoi (argv[2]);

  // ROS init
  init (argc, argv, "object_data_acquisition_gazebo");

  ObjectDataAcquisitionGazebo p (param, model_i);
  p.spin ();

  return (0);
}

